/*
 * forth_lwip_wrapper.c
 * Telnet demo for the Embeddable Forth Command Interpreter.
 * Written by Andras Zsoter.
 * This is a quick and dirty example how to control the Forth Engine over a telnet connection.
 * You can treat the contents of this file as public domain.
 *
 * Attribution is appreciated but not mandatory for the contents of this file.
 *
 */

// http://forth.teleonomix.com/

#include "forth.h"
#include "forth_internal.h"
#include "forth_dict.h"
#include "forth_rtos.h"

#include <xil_printf.h>
// #include <stdio.h>
#include <string.h>


#include "lwip/inet.h"

#include "lwip/sockets.h"
#include "lwip/sys.h"
#include "lwipopts.h"

#ifdef __arm__
#include "xil_printf.h"
#endif
#include <rtos.h>
#include <errno.h>
#include <alloca.h>
#include <stdbool.h>
#undef DEBUG_TELNET

#if defined(RTOS_SMP)
extern void print_cpu(void);
#endif

int socket_error(int sd)
{
	int error = 0;
	u32_t optlen = sizeof(error);
	int ret = lwip_getsockopt(sd, SOL_SOCKET, SO_ERROR, &error, &optlen);
	(void)ret;
	return error;
}

forth_scell_t write_all(int sd, const char *buff, size_t len)
{
	size_t cnt = 0;
	ssize_t written;
	int error;

	while (len > cnt)
	{
#if defined(RTOS_SMP)
//		print_cpu();
#endif
		written = lwip_write(sd, buff + cnt, len - cnt);

		if (0 == written)
		{
			outbyte('+');
			return cnt;
		}

		if (0 > written)
		{
			error = socket_error(sd);
			if ((EAGAIN == error) || (EWOULDBLOCK == error))
			{
				outbyte('#');
				RTOS_Delay(1);
				// vTaskDelay(1);
				// RTOS_Yield();
				continue;
			}

			outbyte('>');
			return written; // Probably -1
		}

		cnt += written;
	}

	return cnt;
}

static int write_str(struct forth_runtime_context *rctx, const char *str, forth_cell_t length)
{
	forth_scell_t written;

#if defined(DEBUG_TELNET)
	char *p = str;
	forth_cell_t len = length;
#endif

	if (0 == length)
	{
		return 0;
	}

#if defined(DEBUG_TELNET)
	while(len--)
	{
		outbyte(*p++);
	}
#endif

	written = write_all(rctx->sd, str, length);

	if (written < 0)
	{
		return -1;
	}

	rctx->terminal_col += length;

	return 0;
}

static int page(struct forth_runtime_context *rctx)
{
	rctx->terminal_col = 0;
	return write_all(rctx->sd, "\014", 1);
}

static int send_cr(struct forth_runtime_context *rctx)
{
	rctx->terminal_col = 0;
	return write_all(rctx->sd, "\r\n", 2);
}

static void emit(struct forth_runtime_context *rctx, char c)
{
	rctx->write_string(rctx, &c, 1);
}

static char *format_num(unsigned int num, char *p)
{
	do
	{
		*--p = '0' + (num % 10);
		num = num / 10;
	} while (0 != num);

	return p;
}

static int at_xy(struct forth_runtime_context *rctx, forth_cell_t x, forth_cell_t y)
{
	char buff[10];
	char *p;
	emit(rctx, 27);								// ESC
	emit(rctx, '[');							// [
	p = format_num(y + 1, buff + 10);
	write_all(rctx->sd, p, (buff + 10) - p);	// column
	emit(rctx,';');								// ;
	p = format_num(x + 1, buff + 10);
	write_all(rctx->sd, p, (buff + 10) - p);	// row
	emit(rctx, 'H');							// H
	rctx->terminal_col = y;
	return 0;
}

// ---------------------------------------------------------------------------------------------------
static forth_scell_t get_next_byte(struct forth_runtime_context *rctx, bool wait)
{
	char c;
	int ret = -1;
	int error;

	while (ret < 0)
	{
		ret = lwip_read(rctx->sd, &c, 1);

		if (0 == ret)
		{
#if defined(DEBUG_TELNET)
			xil_printf("get_next_byte(%d): ret = %d\r\n", rctx->sd, ret);
#endif
			//return 0;
			return -1;
		}

		if (0 > ret)
		{
			error = socket_error(rctx->sd);

#if defined(DEBUG_TELNET)
			xil_printf("get_next_byte(%d): ret = %d error=%d\r\n", rctx->sd, ret, error);
#endif
			if ((EAGAIN == error) || (EWOULDBLOCK == error))
			{
				RTOS_Delay(1);
				continue;
			}
			return -1;
		}
	}

	return (forth_scell_t)c;
}

// --------------------------------------
// Implementing proper telnet negotiation is beyond the scope of this simple example, just throw away telnet stuff.
int telnet_stuff(struct forth_runtime_context *rctx)
{
	forth_cell_t k;

	k = get_next_byte(rctx, true);

	if ((forth_cell_t)-1 == k)
	{
		return -1;
	}

#if defined(DEBUG_TELNET)
	xil_printf("-K = %08x\r\n", k);
#endif

	if (k < 128)
	{
		return 1; // ??
	}

	switch(k)
	{
		case 250: // SB
			do
			{
					do {
						k = get_next_byte(rctx, true);

						if ((forth_cell_t)-1 == k)
						{
							return -1;
						}

#if defined(DEBUG_TELNET)
						xil_printf("-K = %08x\r\n", k);
#endif

					} while (255 != k);	// IAC
					k = get_next_byte(rctx, true);

					if ((forth_cell_t)-1 == k)
					{
						return -1;
					}
#if defined(DEBUG_TELNET)
					xil_printf("-K = %08x\r\n", k);
#endif
			} while(240 != k);	// SE
			return 0;
		break;
		case 251: // WILL
		case 252: // WON'T
		case 253: // DO
		case 254: // DON'T
			k = get_next_byte(rctx, true);

			if ((forth_cell_t)-1 == k)
			{
				return -1;
			}
#if defined(DEBUG_TELNET)
			xil_printf("-K = %08x\r\n", k);
#endif
			return 0;
		break;
	}

	return 0;
}

// Try to get a valid character from the socket.
// If there is some telnet related traffic try to get rid of it.
// Return values:
// -1: Error
//  0: Nothing has happened.
//  1: There is a valid character in rctx->latest_key.
static forth_scell_t try_refill_char(struct forth_runtime_context *rctx)
{
	forth_scell_t ret;

	if (0 < rctx->latest_key)
	{
		return 1; // Already a valid char;
	}

	ret = get_next_byte(rctx, false);

	if (0 > ret)	// Error.
	{
		return -1;
	}

	if (0 == ret)
	{
		return 0;
	}

	if (255 == ret)
	{
		if (0 < telnet_stuff(rctx))	// This could technically block, but we ignore it for KEY and EKEY?
		{
			return -1;
		}
		return 0;
	}

	rctx->latest_key = ret;
	return 1;
}

static forth_scell_t get_next_char(struct forth_runtime_context *rctx)
{
	forth_scell_t ret;

	while (0 > rctx->latest_key)
	{
		if (0 > try_refill_char(rctx))
		{
			return -1;
		}
	}

	ret = rctx->latest_key;
	rctx->latest_key = -1;
	return ret;
}
// ---------------------------------------------------------------------------------------------------
// These are fake, in reality they should be more complicated to follow the rules in the FORTH standard.
static forth_cell_t key(struct forth_runtime_context *rctx)
{


#if 0
	char c;
	int ret;
	ret = lwip_read(rctx->sd, &c, 1);

	if (0 >= ret)
	{
#if defined(DEBUG_TELNET)
		xil_printf("KEY: ret = %d\r\n", ret);
#endif
		return (forth_cell_t)-1;
	}
#else
	forth_scell_t c = get_next_char(rctx);
#endif
    if (127 == c || 8 == c || 7 == c)       // Ugly, but I have no idea what the proper symbolic names are.
    {
            c = '\b';
    }

	return (forth_cell_t)c;
}

// EKEY?
static forth_cell_t ekey_q(struct forth_runtime_context *rctx)
{
	fd_set readfiles;
	struct timeval to = { 0, 0 };
	FD_ZERO(&readfiles);
	FD_SET(rctx->sd, &readfiles);

	if (0 >= lwip_select(rctx->sd + 1, &readfiles, 0, 0, &to))
	{
		return FORTH_FALSE;
	}

	if (0 > try_refill_char(rctx))
	{
		return -1;
	}

	if (0 < rctx->latest_key)
	{
		return 1;
	}

	return FORTH_FALSE;
}

// KEY?
static forth_cell_t key_q(struct forth_runtime_context *rctx)
{
	return rctx->ekey_q(rctx);
}

// EKEY
static forth_cell_t ekey(struct forth_runtime_context *rctx)
{
	forth_cell_t c = rctx->key(rctx);

	if ((forth_cell_t)-1 == c)
	{
		return c;
	}

	return ((forth_cell_t)c) << 8;
}

// EKEY>CHAR
static forth_cell_t ekey_to_char(struct forth_runtime_context *rctx, forth_cell_t ek)
{
	return 0xFF & (ek >> 8);
}

extern void forth_dump(struct forth_runtime_context *rctx, const char *addr, forth_cell_t len);

static forth_scell_t accept_str(struct forth_runtime_context *rctx, char *buffer, forth_cell_t length)
{
	char *p = buffer;
	forth_cell_t k;
	char c;

	for (k = rctx->key(rctx); ('\n' != k) && ('\r' != k) && ((p - buffer) < length); k = rctx->key(rctx))
	{
#if defined(DEBUG_TELNET)
		xil_printf("c = %08x\r\n", k);
#endif
		if ((forth_cell_t)-1 == k)
		{
			return -1;
		}

		c = (char)k;
#if (0)
		if (255 == c)
		{
			if (0 < telnet_stuff(rctx))
			{
				return -1;
			}
			continue;
		}

		if (0 == c)
		{
			continue;
		}
#endif
	    // emit(rctx, c);

	    if ('\b' == c)
	    {
	    	emit(rctx, 0x08);
	    	emit(rctx, ' ');
	    	emit(rctx, 0x08);

	    	if (p > buffer)
	    	{
	    		p--;
	    	}
	    }
	    else
	    {
		    emit(rctx, c);
	    	*p++ = c;
	    }
	}

	emit(rctx, FORTH_CHAR_SPACE);

#if defined(DEBUG_TELNET)
	forth_dump(rctx, buffer, p - buffer);
#endif

	return p - buffer;
}

int forth_run_forth_network(int sd, unsigned int dstack_cells, unsigned int rstack_cells)
{
#define SEARCH_ORDER_SIZE 256 /* Perhaps move this to some header file at one point. */
	struct forth_runtime_context *rctx;
	forth_cell_t *sp;
	forth_cell_t *rp;
	forth_cell_t *search_order;
	forth_cell_t size = sizeof(struct forth_runtime_context) + sizeof(forth_cell_t) * (dstack_cells + rstack_cells + (SEARCH_ORDER_SIZE));
    // char *ctx = malloc(size);
    char *ctx = alloca(size);
    if (0 == ctx)
    {
    	xil_printf("ERROR: Failed to create Forth runtime context!\r\n");
    	return -1;
    }

    memset(ctx, 0, size);

     rctx = (struct forth_runtime_context *)ctx;
     sp = (forth_cell_t *)(rctx + 1);
     rp = sp + dstack_cells;
     search_order = rp + rstack_cells;

    rctx->dictionary = dictionary;
    rctx->sp0 = rp - 1;
    rctx->sp = rctx->sp0;
    rctx->sp_max = rctx->sp0;
   	rctx->sp_min = sp;

   	rctx->rp0 = (rp + rstack_cells) - 1;
   	rctx->rp = rctx->rp0;
   	rctx->rp_max = rctx->rp0;
   	rctx->rp_min = rp;

   	rctx->handler = 0;
   	rctx->ip = 0;
    rctx->base = 10;

    rctx->wordlists = search_order;
   	rctx->wordlist_slots = (SEARCH_ORDER_SIZE);
   	rctx->wordlist_cnt = 2;
   	search_order[(SEARCH_ORDER_SIZE) - 1] = FORTH_WID_Root_WORDLIST;
   	search_order[(SEARCH_ORDER_SIZE) - 2] = FORTH_WID_FORTH_WORDLIST;
    rctx->current = FORTH_WID_FORTH_WORDLIST;
    		// r_ctx.wordlist_cnt =
   	rctx->terminal_width = 80;
   	rctx->terminal_height = 25;
   	rctx->write_string = &write_str;
   	rctx->page = &page;
   	rctx->at_xy = &at_xy;
   	rctx->send_cr = &send_cr;
    rctx->accept_string = &accept_str;
   	rctx->key = &key;
   	rctx->key_q = &key_q;
   	rctx->ekey = &ekey;
   	rctx->ekey_q = &ekey_q;
   	rctx->ekey_to_char = &ekey_to_char;
   	rctx->sd = sd;
   	rctx->latest_key = -1;

#if defined(FORTH_EXTERNAL_PRIMITIVES)
	forth_setup_application_primitives(rctx);
#endif

    forth(rctx, FORTH_XT_QUIT);

//   	free(ctx);
    return 0;
}
